#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-10-11
################################################################

import time
import threading
import os
import h5py
import numpy as np
from collections import deque


class HexHdf5Writer:

    def __init__(
        self,
        file_path: str,
        print_interval: int = 1_000,
        batch_size: int = 64,
    ):
        self.__file_path = file_path
        self.__hdf5_file = h5py.File(file_path, "w", libver='latest')
        self.__group_dict = {}
        self.__dataset_dict = {}
        self.__batch_size = batch_size
        self.__print_num = 0
        self.__print_interval = print_interval

        self.__queue = deque()
        self.__stop_event = threading.Event()
        self.__writer_cnt = 0
        self.__writer_thread = None
        self.__writer_exc = None

    def __del__(self):
        self.stop()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def start(self):
        if self.__writer_thread and self.__writer_thread.is_alive():
            return
        self.__stop_event.clear()
        self.__writer_thread = threading.Thread(
            target=self.__writer_loop,
            daemon=True,
        )
        self.__writer_thread.start()

    def stop(self):
        self.__stop_event.set()
        if self.__writer_thread is not None and self.__writer_thread.is_alive(
        ):
            self.__writer_thread.join()
        self.summary()
        if self.__hdf5_file is not None:
            try:
                self.__hdf5_file.flush()
                self.__hdf5_file.close()
            except Exception:
                pass
            self.__hdf5_file = None
        if self.__writer_exc:
            raise self.__writer_exc

    def get_shape(self, group_name: str, dataset_name: str = "data"):
        dataset = self.__get_dataset_handle(group_name, dataset_name)
        return dataset.shape

    def get_dtype(self, group_name: str, dataset_name: str = "data"):
        dataset = self.__get_dataset_handle(group_name, dataset_name)
        return dataset.dtype

    def __get_dataset_handle(
        self,
        group_name: str,
        dataset_name: str = "data",
    ):
        dataset_key = f"{group_name}/{dataset_name}"
        if dataset_key not in self.__dataset_dict:
            if group_name not in self.__hdf5_file:
                raise KeyError(f"Group '{group_name}' not found in HDF5 file")
            if dataset_name not in self.__hdf5_file[group_name]:
                raise KeyError(
                    f"Dataset '{dataset_name}' not found in group '{group_name}'"
                )
            self.__dataset_dict[dataset_key] = self.__hdf5_file[group_name][
                dataset_name]
        return self.__dataset_dict[dataset_key]

    def summary(self):
        if self.__hdf5_file is None:
            print("HDF5 file is closed. Cannot get summary.")
            return

        print("#" * 50)
        print(f"HDF5 File: {self.__file_path}")
        print("#" * 50)

        for group_name in self.__hdf5_file.keys():
            print("-" * 30)
            print(f"Group: {group_name}")
            print("-" * 30)
            get_delta_s = ((self.__hdf5_file[group_name]['get_ts'][-1] -
                            self.__hdf5_file[group_name]['get_ts'][0]) *
                           1e-9)[0]
            sen_delta_s = ((self.__hdf5_file[group_name]['sen_ts'][-1] -
                            self.__hdf5_file[group_name]['sen_ts'][0]) *
                           1e-9)[0]
            print(f"get_delta: {get_delta_s}s")
            print(f"sen_delta: {sen_delta_s}s")

            dtype = self.get_dtype(group_name)
            shape = self.get_shape(group_name)
            print(f"  Dataset: {group_name}")
            print(f"    Dtype: {dtype}")
            print(f"    Shape: {shape}")
            print(f"    GetHz: {shape[0] / get_delta_s}")
            print(f"    SenHz: {shape[0] / sen_delta_s}")

        print("#" * 50)

    def __writer_loop(self):
        try:
            while not self.__stop_event.is_set():
                # Collect batch of items
                batch_items = []
                try:
                    # Try to collect up to batch_size items
                    for _ in range(self.__batch_size):
                        item = self.__queue.popleft()
                        batch_items.append(item)
                except IndexError:
                    # Queue is empty or not enough items
                    if len(batch_items) == 0:
                        time.sleep(1e-5)
                        continue

                # Write batch if we have items
                if batch_items:
                    self.__write_batch(batch_items)

            # Flush remaining items
            empty_count = 0
            max_empty_checks = 10
            while empty_count < max_empty_checks:
                batch_items = []
                try:
                    # Collect remaining items in batches
                    for _ in range(self.__batch_size):
                        item = self.__queue.popleft()
                        batch_items.append(item)
                    empty_count = 0
                except IndexError:
                    # Queue is empty, but we may have collected some items
                    # Write them before continuing
                    if batch_items:
                        self.__write_batch(batch_items)
                        batch_items = []
                    empty_count += 1
                    if empty_count >= max_empty_checks:
                        break
                    time.sleep(1e-5)
                    continue

                # Write batch if we have items
                if batch_items:
                    self.__write_batch(batch_items)
        except Exception as e:
            self.__writer_exc = e
            raise

    def __write_batch(self, batch_items):
        """Write a batch of items, grouped by group_name for efficiency."""
        # Group items by group_name
        grouped_items = {}
        for item in batch_items:
            group = item[0]
            if group not in grouped_items:
                grouped_items[group] = []
            grouped_items[group].append(item)

        # Write each group's batch
        for group, items in grouped_items.items():
            self.__write_group_batch(group, items)

    def __write_group_batch(self, group_name, items):
        """Write a batch of items for a specific group."""
        dataset_key = f"{group_name}/data"
        get_ts_key = f"{group_name}/get_ts"
        sen_ts_key = f"{group_name}/sen_ts"
        ds = self.__dataset_dict[dataset_key]
        d_get = self.__dataset_dict[get_ts_key]
        d_sen = self.__dataset_dict[sen_ts_key]

        batch_size = len(items)
        n_old = ds.shape[0]
        n_new = n_old + batch_size

        # Resize all datasets once
        ds.resize((n_new, *ds.shape[1:]))
        d_get.resize((n_new, 1))
        d_sen.resize((n_new, 1))

        # Prepare batch arrays
        data_list = []
        gts_list = []
        sts_list = []

        for group, data, gts, sts in items:
            data_list.append(data)
            # Ensure gts and sts are scalars or 1-element arrays
            if isinstance(gts, np.ndarray):
                gts_val = gts.item() if gts.size == 1 else gts[0]
            else:
                gts_val = int(gts)
            if isinstance(sts, np.ndarray):
                sts_val = sts.item() if sts.size == 1 else sts[0]
            else:
                sts_val = int(sts)
            gts_list.append(gts_val)
            sts_list.append(sts_val)

        # Stack arrays for batch write
        data_batch = np.stack(data_list, axis=0)
        gts_batch = np.array(gts_list, dtype=np.int64).reshape(-1, 1)
        sts_batch = np.array(sts_list, dtype=np.int64).reshape(-1, 1)

        # Batch write
        ds[n_old:n_new, ...] = data_batch
        d_get[n_old:n_new, :] = gts_batch
        d_sen[n_old:n_new, :] = sts_batch

        self.__writer_cnt += batch_size
        cur_print_num = self.__writer_cnt // self.__print_interval
        if cur_print_num > self.__print_num:
            self.__print_num = cur_print_num
            print("#" * 50)
            for group_name in self.__hdf5_file.keys():
                print(f"{group_name} len:{self.get_shape(group_name)[0]}")

    def create_dataset(
        self,
        group_name: str,
        shape: tuple,
        dtype: np.dtype,
        chunk_num: int,
        max_num: int | None = None,
        compression=None,
    ):
        if group_name not in self.__group_dict:
            self.__group_dict[group_name] = self.__hdf5_file.create_group(
                group_name)

        dataset = self.__group_dict[group_name].create_dataset(
            "data",
            shape=(0, *shape),
            maxshape=(max_num, *shape),
            dtype=dtype,
            chunks=(chunk_num, *shape),
            compression=compression,
        )
        get_ts_set = self.__group_dict[group_name].create_dataset(
            "get_ts",
            shape=(0, 1),
            maxshape=(max_num, 1),
            dtype=np.int64,
            chunks=(chunk_num, 1),
        )
        sen_ts_set = self.__group_dict[group_name].create_dataset(
            "sen_ts",
            shape=(0, 1),
            maxshape=(max_num, 1),
            dtype=np.int64,
            chunks=(chunk_num, 1),
        )

        # Store dataset reference for easy access
        self.__dataset_dict[f"{group_name}/data"] = dataset
        self.__dataset_dict[f"{group_name}/get_ts"] = get_ts_set
        self.__dataset_dict[f"{group_name}/sen_ts"] = sen_ts_set

    def append_data(
        self,
        group_name: str,
        data: np.ndarray,
        get_ts: np.ndarray | int,
        sen_ts: np.ndarray | int,
    ):
        if isinstance(get_ts, int):
            get_ts = np.array([get_ts])
        if isinstance(sen_ts, int):
            sen_ts = np.array([sen_ts])
        item = (
            group_name,
            data,
            get_ts,
            sen_ts,
        )
        self.__queue.append(item)

    def append_batch_data(
        self,
        group_name: str,
        data: np.ndarray,
        get_ts: np.ndarray,
        sen_ts: np.ndarray,
    ):
        """Append batch data more efficiently by adding all items to queue at once."""
        batch_size = data.shape[0]
        # Ensure get_ts and sen_ts are properly shaped
        if get_ts.ndim == 0:
            get_ts = np.array([get_ts] * batch_size)
        elif get_ts.shape[0] != batch_size:
            raise ValueError(
                f"get_ts shape mismatch: expected {batch_size}, got {get_ts.shape[0]}"
            )

        if sen_ts.ndim == 0:
            sen_ts = np.array([sen_ts] * batch_size)
        elif sen_ts.shape[0] != batch_size:
            raise ValueError(
                f"sen_ts shape mismatch: expected {batch_size}, got {sen_ts.shape[0]}"
            )

        # Add all items to queue efficiently
        for i in range(batch_size):
            item = (
                group_name,
                data[i],
                get_ts[i] if isinstance(get_ts[i], np.ndarray) else np.array(
                    [get_ts[i]]),
                sen_ts[i] if isinstance(sen_ts[i], np.ndarray) else np.array(
                    [sen_ts[i]]),
            )
            self.__queue.append(item)


class HexHdf5MultiWriter:

    def __init__(self, base_dir: str):
        os.makedirs(base_dir, exist_ok=True)
        arm_path = f"{base_dir}/arm.h5"
        rgb_path = f"{base_dir}/rgb.h5"
        depth_path = f"{base_dir}/depth.h5"
        self.__writers: dict[str, HexHdf5Writer] = {
            "arm": HexHdf5Writer(arm_path, 10_000, batch_size=1024),
            "rgb": HexHdf5Writer(rgb_path, 300, batch_size=4),
            "depth": HexHdf5Writer(depth_path, 300, batch_size=4),
        }

    def start(self):
        for key, writer in self.__writers.items():
            print(f"Starting writer for {key}")
            writer.start()

    def stop(self):
        for key, writer in self.__writers.items():
            print(f"Stopping writer for {key}")
            writer.stop()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    # --------------- proxies ---------------
    def get_writer(self, msg_type: str) -> HexHdf5Writer:
        return self.__writers[msg_type]

    def create_dataset(
        self,
        msg_type: str,
        group_name: str,
        shape: tuple,
        dtype: np.dtype,
        chunk_num: int,
        max_num: int | None = None,
    ):
        self.__writers[msg_type].create_dataset(
            group_name=group_name,
            shape=shape,
            dtype=dtype,
            chunk_num=chunk_num,
            max_num=max_num,
            compression=None,
        )

    def append_data(
        self,
        msg_type: str,
        group_name: str,
        data: np.ndarray,
        get_ts: np.ndarray | int,
        sen_ts: np.ndarray | int,
    ):
        self.__writers[msg_type].append_data(
            group_name=group_name,
            data=data,
            get_ts=get_ts,
            sen_ts=sen_ts,
        )

    def append_batch_data(
        self,
        msg_type: str,
        group_name: str,
        data: np.ndarray,
        get_ts: np.ndarray,
        sen_ts: np.ndarray,
    ):
        self.__writers[msg_type].append_batch_data(
            group_name=group_name,
            data=data,
            get_ts=get_ts,
            sen_ts=sen_ts,
        )

    def now_ns(self):
        return np.array([time.time_ns()])

    def hex_ts_to_ns(self, ts: dict):
        try:
            return np.array([ts["s"] * 1e9 + ts["ns"]])
        except Exception as e:
            print(f"hex_ts_to_ns failed: {e}")
            return np.array([np.inf])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-11-13
################################################################

import numpy as np


class HexPlanUtilBvp:

    def __init__(
        self,
        dt,
        jnt_num: int = 6,
        tar_num: int = 2,
        adjust_weights: np.ndarray = None,
    ):
        # q = a0 + a1*t + a2*t^2 + a3*t^3
        self.__jnt_num = jnt_num
        self.__tar_num = tar_num
        self.__dt_arr = np.arange(self.__tar_num) * dt + dt
        self.__dt2_arr = self.__dt_arr**2
        self.__dt3_arr = self.__dt_arr**3
        self.__a_mat = np.vstack((self.__dt2_arr, self.__dt3_arr)).T

        h_mat = self.__a_mat @ np.linalg.inv(
            self.__a_mat.T @ self.__a_mat) @ self.__a_mat.T
        leverage = np.diag(h_mat)
        self.__weights_leverage = 1.0 / (leverage + 1e-6)

        if adjust_weights is not None:
            assert adjust_weights.shape == (self.__tar_num, )
            assert np.all(
                adjust_weights >= 0), "adjust weights must be non-negative"
            self.__weights_leverage = self.__weights_leverage * adjust_weights

        self.__weights_leverage = self.__weights_leverage / np.linalg.norm(
            self.__weights_leverage)

    def __call__(
        self,
        q0: np.ndarray,
        dq0: np.ndarray,
        q_tar_arr: np.ndarray,
    ):
        assert q0.shape == (self.__jnt_num, )
        assert dq0.shape == (self.__jnt_num, )
        assert q_tar_arr.shape == (self.__tar_num, self.__jnt_num)

        a0_arr = q0.copy()
        a1_arr = dq0.copy()

        y = q_tar_arr - (a0_arr[None, :] +
                         a1_arr[None, :] * self.__dt_arr[:, None])

        # weights for least squares
        sqrt_weights = np.sqrt(self.__weights_leverage)
        a_mat_weighted = self.__a_mat * sqrt_weights[:, None]
        y_weighted = y * sqrt_weights[:, None]

        # least squares
        sol, residuals, ranks, svals = np.linalg.lstsq(
            a_mat_weighted,
            y_weighted,
            rcond=None,
        )
        a2_arr = sol[0, :]
        a3_arr = sol[1, :]

        if residuals.ndim == 0:
            residuals_arr = np.full(self.__jnt_num, residuals)
        else:
            residuals_arr = residuals

        return {
            "format": "a0 + a1*t + a2*t^2 + a3*t^3",
            "a0": a0_arr,
            "a1": a1_arr,
            "a2": a2_arr,
            "a3": a3_arr,
            "residuals": residuals_arr,
            "ranks": ranks,
            "singular_values": svals,
        }

    def calc_state(self, dt: float, params: dict):
        a0 = params["a0"]
        a1 = params["a1"]
        a2 = params["a2"]
        a3 = params["a3"]
        q = a0 + a1 * dt + a2 * dt**2 + a3 * dt**3
        dq = a1 + 2 * a2 * dt + 3 * a3 * dt**2
        ddq = 2 * a2 + 6 * a3 * dt
        return {
            "q": q,
            "dq": dq,
            "ddq": ddq,
        }



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-01-14
################################################################

import numpy as np
from typing import Tuple


def hat(vec: np.ndarray):
    """so(3) vector → skew-symmetric matrix"""
    assert (vec.ndim == 1 and vec.shape == (3, )) or (
        vec.ndim == 2 and vec.shape[1] == 3), "cross_matrix vec shape err"

    if vec.ndim == 1:
        mat = np.array([
            [0.0, -vec[2], vec[1]],
            [vec[2], 0.0, -vec[0]],
            [-vec[1], vec[0], 0.0],
        ])
    else:
        # Batch processing: input shape (N, 3) -> output shape (N, 3, 3)
        num = vec.shape[0]
        mat = np.zeros((num, 3, 3))
        mat[:, 0, 1] = -vec[:, 2]
        mat[:, 0, 2] = vec[:, 1]
        mat[:, 1, 0] = vec[:, 2]
        mat[:, 1, 2] = -vec[:, 0]
        mat[:, 2, 0] = -vec[:, 1]
        mat[:, 2, 1] = vec[:, 0]
    return mat


def vee(mat: np.ndarray):
    """skew-symmetric matrix → so(3) vector"""
    assert (mat.ndim == 2 and mat.shape == (3, 3)) or (
        mat.ndim == 3 and mat.shape[1:] == (3, 3)), "vee mat shape err"

    if mat.ndim == 2:
        vec = np.array([mat[2, 1], mat[0, 2], mat[1, 0]])
    else:
        # Batch processing: input shape (N, 3, 3) -> output shape (N, 3)
        vec = np.zeros((mat.shape[0], 3))
        vec[:, 0] = mat[:, 2, 1]
        vec[:, 1] = mat[:, 0, 2]
        vec[:, 2] = mat[:, 1, 0]
    return vec


def rad2deg(rad):
    deg = rad * 180.0 / np.pi
    return deg


def deg2rad(deg):
    rad = deg * np.pi / 180.0
    return rad


def angle_norm(rad):
    normed_rad = (rad + np.pi) % (2 * np.pi) - np.pi
    return normed_rad


def vec_norm(vec: np.ndarray) -> np.ndarray:
    assert (vec.ndim == 1 and vec.shape == (3, )) or (
        vec.ndim == 2 and vec.shape[1] == 3), "vec_norm vec shape err"

    single = vec.ndim == 1
    if single:
        vec = vec[np.newaxis, :]

    normed_vec = vec / np.linalg.norm(vec, axis=1, keepdims=True)
    return normed_vec[0] if single else normed_vec


def quat_norm(quat: np.ndarray) -> np.ndarray:
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat_norm quat shape err"

    single = quat.ndim == 1
    if single:
        quat = quat[np.newaxis, :]

    # Normalize quaternion
    normed_quat = quat / np.linalg.norm(quat, axis=1, keepdims=True)

    # Ensure first component (w) is non-negative using vectorized operation
    sign = np.where(normed_quat[:, 0:1] < 0.0, -1.0, 1.0)
    normed_quat = normed_quat * sign

    return normed_quat[0] if single else normed_quat


def quat_slerp(q1: np.ndarray, q2: np.ndarray, t) -> np.ndarray:
    assert ((q1.ndim == 1 and q1.shape == (4, ))
            or (q1.ndim == 2 and q1.shape[1] == 4)), "quat_slerp q1 shape err"
    assert ((q2.ndim == 1 and q2.shape == (4, ))
            or (q2.ndim == 2 and q2.shape[1] == 4)), "quat_slerp q2 shape err"
    assert q1.ndim == q2.ndim, "quat_slerp q1 and q2 must have same ndim"

    single = q1.ndim == 1
    if single:
        assert isinstance(
            t, (int, float)), "quat_slerp t must be scalar for 1D quat"
        q1 = q1[np.newaxis, :]
        q2 = q2[np.newaxis, :]
        t = np.array([t])
    else:
        assert q1.shape[0] == q2.shape[0], "quat_slerp batch size mismatch"
        assert (isinstance(t, np.ndarray) and t.ndim == 1 and
                t.shape[0] == q1.shape[0]) or isinstance(t, (int, float)), \
            "quat_slerp t shape err"
        if isinstance(t, (int, float)):
            t = np.full(q1.shape[0], t)

    # Normalize using quat_norm
    q1_norm = quat_norm(q1)
    q2_norm = quat_norm(q2)

    # dot
    dot = np.sum(q1_norm * q2_norm, axis=1)
    neg_dot_mask = dot < 0.0
    q2_norm[neg_dot_mask] = -q2_norm[neg_dot_mask]
    dot[neg_dot_mask] = -dot[neg_dot_mask]
    dot = np.clip(dot, -1.0, 1.0)
    theta = np.arccos(dot)

    # slerp
    small_mask = np.fabs(theta) < 1e-6
    q = np.zeros_like(q1)

    if np.any(small_mask):
        q[small_mask] = q1_norm[small_mask] + t[small_mask, np.newaxis] * (
            q2_norm[small_mask] - q1_norm[small_mask])
        q[small_mask] = quat_norm(q[small_mask])

    if np.any(~small_mask):
        sin_theta = np.sin(theta[~small_mask])
        q1_factor = np.sin(
            (1 - t[~small_mask]) * theta[~small_mask]) / sin_theta
        q2_factor = np.sin(t[~small_mask] * theta[~small_mask]) / sin_theta
        q[~small_mask] = (q1_factor[:, np.newaxis] * q1_norm[~small_mask] +
                          q2_factor[:, np.newaxis] * q2_norm[~small_mask])

    return q[0] if single else q


def quat_mul(q1: np.ndarray, q2: np.ndarray) -> np.ndarray:
    assert ((q1.ndim == 1 and q1.shape == (4, ))
            or (q1.ndim == 2 and q1.shape[1] == 4)), "quat_mul q1 shape err"
    assert ((q2.ndim == 1 and q2.shape == (4, ))
            or (q2.ndim == 2 and q2.shape[1] == 4)), "quat_mul q2 shape err"
    assert q1.ndim == q2.ndim, "quat_mul q1 and q2 must have same ndim"

    single = q1.ndim == 1
    if single:
        q1 = q1[np.newaxis, :]
        q2 = q2[np.newaxis, :]
    else:
        assert q1.shape[0] == q2.shape[0], "quat_mul batch size mismatch"

    # Normalize using quat_norm
    q1_norm = quat_norm(q1)
    q2_norm = quat_norm(q2)

    # mul
    w1, x1, y1, z1 = q1_norm[:, 0], q1_norm[:, 1], q1_norm[:, 2], q1_norm[:, 3]
    w2, x2, y2, z2 = q2_norm[:, 0], q2_norm[:, 1], q2_norm[:, 2], q2_norm[:, 3]
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 - x1 * z2 + y1 * w2 + z1 * x2
    z = w1 * z2 + x1 * y2 - y1 * x2 + z1 * w2
    q = np.stack([w, x, y, z], axis=1)

    return q[0] if single else q


def quat_inv(quat: np.ndarray) -> np.ndarray:
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat_inv quat shape err"

    single = quat.ndim == 1
    if single:
        quat = quat[np.newaxis, :]

    # Normalize using quat_norm
    q = quat_norm(quat)

    # inv
    inv = np.stack([q[:, 0], -q[:, 1], -q[:, 2], -q[:, 3]], axis=1)

    return inv[0] if single else inv


def trans_inv(trans: np.ndarray) -> np.ndarray:
    assert (trans.ndim == 2 and trans.shape == (4, 4)) or (
        trans.ndim == 3
        and trans.shape[1:] == (4, 4)), "trans_inv trans shape err"

    single = trans.ndim == 2
    if single:
        trans = trans[np.newaxis, ...]

    pos = trans[:, :3, 3]
    rot = trans[:, :3, :3]

    # inv
    inv = np.zeros_like(trans)
    inv[:, :3, :3] = rot.transpose(0, 2, 1)
    inv[:, :3, 3] = -np.einsum('ijk,ik->ij', inv[:, :3, :3], pos)
    inv[:, 3, :3] = 0.0
    inv[:, 3, 3] = 1.0

    return inv[0] if single else inv


def rot2quat(rot: np.ndarray) -> np.ndarray:
    assert (rot.ndim == 2 and rot.shape == (3, 3)) or (
        rot.ndim == 3 and rot.shape[1:] == (3, 3)), "rot2quat rot shape err"

    single = rot.ndim == 2
    if single:
        rot = rot[np.newaxis, ...]

    # Batch processing: input shape (N, 3, 3) -> output shape (N, 4)
    num = rot.shape[0]
    quat = np.zeros((num, 4))
    trace = np.trace(rot, axis1=1, axis2=2)

    # Case 1: trace > 0
    mask1 = trace > 0
    if np.any(mask1):
        temp = 2.0 * np.sqrt(1 + trace[mask1])
        quat[mask1, 0] = 0.25 * temp
        quat[mask1, 1] = (rot[mask1, 2, 1] - rot[mask1, 1, 2]) / temp
        quat[mask1, 2] = (rot[mask1, 0, 2] - rot[mask1, 2, 0]) / temp
        quat[mask1, 3] = (rot[mask1, 1, 0] - rot[mask1, 0, 1]) / temp

    # Case 2: trace <= 0
    mask2 = ~mask1
    if np.any(mask2):
        diag = np.diagonal(rot, axis1=1, axis2=2)
        mask2a = mask2 & (diag[:, 0] > diag[:, 1]) & (diag[:, 0] > diag[:, 2])
        mask2b = mask2 & (~mask2a) & (diag[:, 1] > diag[:, 2])
        mask2c = mask2 & (~mask2a) & (~mask2b)

        # Case 2a: rot[0,0] > rot[1,1] and rot[0,0] > rot[2,2]
        if np.any(mask2a):
            temp = 2.0 * np.sqrt(1 + diag[mask2a, 0] - diag[mask2a, 1] -
                                 diag[mask2a, 2])
            quat[mask2a, 0] = (rot[mask2a, 2, 1] - rot[mask2a, 1, 2]) / temp
            quat[mask2a, 1] = 0.25 * temp
            quat[mask2a, 2] = (rot[mask2a, 1, 0] + rot[mask2a, 0, 1]) / temp
            quat[mask2a, 3] = (rot[mask2a, 0, 2] + rot[mask2a, 2, 0]) / temp

        # Case 2b: rot[1,1] > rot[2,2]
        if np.any(mask2b):
            temp = 2.0 * np.sqrt(1 + diag[mask2b, 1] - diag[mask2b, 0] -
                                 diag[mask2b, 2])
            quat[mask2b, 0] = (rot[mask2b, 0, 2] - rot[mask2b, 2, 0]) / temp
            quat[mask2b, 1] = (rot[mask2b, 1, 0] + rot[mask2b, 0, 1]) / temp
            quat[mask2b, 2] = 0.25 * temp
            quat[mask2b, 3] = (rot[mask2b, 2, 1] + rot[mask2b, 1, 2]) / temp

        # Case 2c: rot[2,2] >= rot[0,0] and rot[2,2] >= rot[1,1]
        if np.any(mask2c):
            temp = 2.0 * np.sqrt(1 + diag[mask2c, 2] - diag[mask2c, 0] -
                                 diag[mask2c, 1])
            quat[mask2c, 0] = (rot[mask2c, 1, 0] - rot[mask2c, 0, 1]) / temp
            quat[mask2c, 1] = (rot[mask2c, 0, 2] + rot[mask2c, 2, 0]) / temp
            quat[mask2c, 2] = (rot[mask2c, 2, 1] + rot[mask2c, 1, 2]) / temp
            quat[mask2c, 3] = 0.25 * temp

    return quat[0] if single else quat


def rot2axis(rot: np.ndarray) -> Tuple[np.ndarray, float]:
    assert (rot.ndim == 2 and rot.shape == (3, 3)) or (
        rot.ndim == 3 and rot.shape[1:] == (3, 3)), "rot2axis rot shape err"

    single = rot.ndim == 2
    if single:
        rot = rot[np.newaxis, ...]

    cos_theta = 0.5 * (np.trace(rot, axis1=1, axis2=2) - 1)
    cos_theta = np.clip(cos_theta, -1.0, 1.0)
    theta = np.arccos(cos_theta)

    small_mask = theta < 1e-6
    axis = np.zeros((rot.shape[0], 3))
    axis[small_mask] = np.array([1.0, 0.0, 0.0])

    if np.any(~small_mask):
        axis_matrix = (rot[~small_mask] - rot[~small_mask].transpose(
            0, 2, 1)) / (2 *
                         np.sin(theta[~small_mask])[:, np.newaxis, np.newaxis])
        axis[~small_mask] = vee(axis_matrix)

    return (axis[0], theta[0]) if single else (axis, theta)


def rot2so3(rot: np.ndarray) -> np.ndarray:
    assert (rot.ndim == 2 and rot.shape == (3, 3)) or (
        rot.ndim == 3 and rot.shape[1:] == (3, 3)), "rot2so3 rot shape err"

    single = rot.ndim == 2
    axis, theta = rot2axis(rot)
    return theta * axis if single else theta[:, np.newaxis] * axis


def quat2rot(quat: np.ndarray) -> np.ndarray:
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat2rot quat shape err"

    single = quat.ndim == 1
    if single:
        quat = quat[np.newaxis, :]

    # Normalize using quat_norm
    q = quat_norm(quat)

    # temp vars
    qx2 = q[:, 1] * q[:, 1]
    qy2 = q[:, 2] * q[:, 2]
    qz2 = q[:, 3] * q[:, 3]
    qxqw = q[:, 1] * q[:, 0]
    qyqw = q[:, 2] * q[:, 0]
    qzqw = q[:, 3] * q[:, 0]
    qxqy = q[:, 1] * q[:, 2]
    qyqz = q[:, 2] * q[:, 3]
    qzqx = q[:, 3] * q[:, 1]

    # rot
    num = q.shape[0]
    rot = np.zeros((num, 3, 3))
    rot[:, 0, 0] = 1 - 2 * (qy2 + qz2)
    rot[:, 0, 1] = 2 * (qxqy - qzqw)
    rot[:, 0, 2] = 2 * (qzqx + qyqw)
    rot[:, 1, 0] = 2 * (qxqy + qzqw)
    rot[:, 1, 1] = 1 - 2 * (qx2 + qz2)
    rot[:, 1, 2] = 2 * (qyqz - qxqw)
    rot[:, 2, 0] = 2 * (qzqx - qyqw)
    rot[:, 2, 1] = 2 * (qyqz + qxqw)
    rot[:, 2, 2] = 1 - 2 * (qx2 + qy2)

    return rot[0] if single else rot


def quat2axis(quat: np.ndarray) -> Tuple[np.ndarray, float]:
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat2axis quat shape err"

    single = quat.ndim == 1
    if single:
        quat = quat[np.newaxis, :]

    # Normalize using quat_norm
    q = quat_norm(quat)

    vec = q[:, 1:]
    norm_vec = np.linalg.norm(vec, axis=1)
    small_mask = norm_vec < 1e-6

    axis = np.zeros((q.shape[0], 3))
    axis[~small_mask] = vec_norm(vec[~small_mask])
    axis[small_mask] = np.array([1.0, 0.0, 0.0])

    theta = np.zeros(q.shape[0])
    theta[~small_mask] = 2 * np.arctan2(norm_vec[~small_mask], q[~small_mask,
                                                                 0])

    return (axis[0], theta[0]) if single else (axis, theta)


def quat2so3(quat: np.ndarray) -> np.ndarray:
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat2so3 quat shape err"

    single = quat.ndim == 1
    axis, theta = quat2axis(quat)

    return theta * axis if single else theta[:, np.newaxis] * axis


def axis2rot(axis: np.ndarray, theta) -> np.ndarray:
    assert (axis.ndim == 1 and axis.shape == (3, )) or (
        axis.ndim == 2 and axis.shape[1] == 3), "axis2rot axis shape err"

    single = axis.ndim == 1
    if single:
        assert isinstance(
            theta, (int, float)), "axis2rot theta must be scalar for 1D axis"
        axis = axis[np.newaxis, :]
        theta = np.array([theta])
    else:
        assert (isinstance(theta, np.ndarray) and theta.ndim == 1 and
                theta.shape[0] == axis.shape[0]) or isinstance(theta, (int, float)), \
            "axis2rot theta shape err"
        if isinstance(theta, (int, float)):
            theta = np.full(axis.shape[0], theta)

    # Normalize axis using vec_norm
    normed_axis = vec_norm(axis)

    num = axis.shape[0]
    rot = np.zeros((num, 3, 3))
    small_mask = theta < 1e-6
    rot[small_mask] = np.eye(3)

    if np.any(~small_mask):
        axis_matrix = hat(normed_axis[~small_mask])
        sin_theta = np.sin(theta[~small_mask])
        cos_theta = np.cos(theta[~small_mask])
        # Batch matrix multiplication
        axis_matrix_sq = axis_matrix @ axis_matrix
        rot[~small_mask] = (
            np.eye(3)[np.newaxis, :, :] +
            sin_theta[:, np.newaxis, np.newaxis] * axis_matrix +
            (1 - cos_theta)[:, np.newaxis, np.newaxis] * axis_matrix_sq)

    return rot[0] if single else rot


def axis2quat(axis: np.ndarray, theta) -> np.ndarray:
    assert (axis.ndim == 1 and axis.shape == (3, )) or (
        axis.ndim == 2 and axis.shape[1] == 3), "axis2quat axis shape err"

    single = axis.ndim == 1
    if single:
        assert isinstance(
            theta, (int, float)), "axis2quat theta must be scalar for 1D axis"
        axis = axis[np.newaxis, :]
        theta = np.array([theta])
    else:
        assert (isinstance(theta, np.ndarray) and theta.ndim == 1 and
                theta.shape[0] == axis.shape[0]) or isinstance(theta, (int, float)), \
            "axis2quat theta shape err"
        if isinstance(theta, (int, float)):
            theta = np.full(axis.shape[0], theta)

    # Normalize axis using vec_norm
    normed_axis = vec_norm(axis)

    num = axis.shape[0]
    quat = np.zeros((num, 4))
    small_mask = theta < 1e-6
    quat[small_mask] = np.array([1.0, 0.0, 0.0, 0.0])

    if np.any(~small_mask):
        half_theta = theta[~small_mask] / 2
        quat[~small_mask, 0] = np.cos(half_theta)
        quat[~small_mask,
             1:] = normed_axis[~small_mask] * np.sin(half_theta)[:, np.newaxis]

    return quat[0] if single else quat


def axis2so3(axis: np.ndarray, theta) -> np.ndarray:
    assert (axis.ndim == 1 and axis.shape == (3, )) or (
        axis.ndim == 2 and axis.shape[1] == 3), "axis2so3 axis shape err"

    single = axis.ndim == 1
    if single:
        assert isinstance(
            theta, (int, float)), "axis2so3 theta must be scalar for 1D axis"
        axis = axis[np.newaxis, :]
        theta = np.array([theta])
    else:
        assert (isinstance(theta, np.ndarray) and theta.ndim == 1 and
                theta.shape[0] == axis.shape[0]) or isinstance(theta, (int, float)), \
            "axis2so3 theta shape err"
        if isinstance(theta, (int, float)):
            theta = np.full(axis.shape[0], theta)

    normed_axis = vec_norm(axis)
    so3 = theta[:, np.newaxis] * normed_axis
    return so3[0] if single else so3


def so32rot(so3: np.ndarray) -> np.ndarray:
    assert (so3.ndim == 1 and so3.shape == (3, )) or (
        so3.ndim == 2 and so3.shape[1] == 3), "so32rot so3 shape err"

    single = so3.ndim == 1
    if single:
        so3 = so3[np.newaxis, :]

    theta = np.linalg.norm(so3, axis=1)
    small_mask = theta < 1e-6

    # Normalize axis using vec_norm
    axis = np.zeros_like(so3)
    axis[~small_mask] = vec_norm(so3[~small_mask])
    axis[small_mask] = np.array([1.0, 0.0, 0.0])

    return axis2rot(axis, theta)


def so32quat(so3: np.ndarray) -> np.ndarray:
    assert (so3.ndim == 1 and so3.shape == (3, )) or (
        so3.ndim == 2 and so3.shape[1] == 3), "so32quat so3 shape err"

    axis, theta = so32axis(so3)
    return axis2quat(axis, theta)


def so32axis(so3: np.ndarray) -> Tuple[np.ndarray, float]:
    assert (so3.ndim == 1 and so3.shape == (3, )) or (
        so3.ndim == 2 and so3.shape[1] == 3), "so32axis so3 shape err"

    single = so3.ndim == 1
    if single:
        so3 = so3[np.newaxis, :]

    theta = np.linalg.norm(so3, axis=1)
    small_mask = theta < 1e-6

    # Normalize axis using vec_norm
    axis = np.zeros((so3.shape[0], 3))
    axis[~small_mask] = vec_norm(so3[~small_mask])
    axis[small_mask] = np.array([1.0, 0.0, 0.0])

    return (axis[0], theta[0]) if single else (axis, theta)


def trans2part(trans: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    assert (trans.ndim == 2 and trans.shape == (4, 4)) or (
        trans.ndim == 3
        and trans.shape[1:] == (4, 4)), "trans2part trans shape err"

    single = trans.ndim == 2
    if single:
        trans = trans[np.newaxis, ...]

    # Batch processing: input shape (N, 4, 4) -> output (N, 3), (N, 4)
    pos = trans[:, :3, 3]
    quat = rot2quat(trans[:, :3, :3])

    return (pos[0], quat[0]) if single else (pos, quat)


def trans2se3(trans: np.ndarray) -> np.ndarray:
    assert (trans.ndim == 2 and trans.shape == (4, 4)) or (
        trans.ndim == 3
        and trans.shape[1:] == (4, 4)), "trans2se3 trans shape err"

    single = trans.ndim == 2
    if single:
        trans = trans[np.newaxis, ...]

    # Batch processing: input shape (N, 4, 4) -> output shape (N, 6)
    pos = trans[:, :3, 3]
    so3 = rot2so3(trans[:, :3, :3])
    se3 = np.concatenate((pos, so3), axis=1)
    return se3[0] if single else se3


def part2trans(pos: np.ndarray, quat: np.ndarray) -> np.ndarray:
    assert (pos.ndim == 1 and pos.shape == (3, )) or (
        pos.ndim == 2 and pos.shape[1] == 3), "part2trans pos shape err"
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "part2trans quat shape err"
    assert pos.ndim == quat.ndim, "part2trans pos and quat must have same ndim"
    if pos.ndim == 2:
        assert pos.shape[0] == quat.shape[0], "part2trans batch size mismatch"

    single = pos.ndim == 1
    if single:
        pos = pos[np.newaxis, :]
        quat = quat[np.newaxis, :]

    num = pos.shape[0]
    trans = np.zeros((num, 4, 4))
    trans[:, :3, 3] = pos
    trans[:, :3, :3] = quat2rot(quat)
    trans[:, 3, :3] = 0.0
    trans[:, 3, 3] = 1.0
    return trans[0] if single else trans


def part2se3(pos: np.ndarray, quat: np.ndarray) -> np.ndarray:
    assert (pos.ndim == 1 and pos.shape == (3, )) or (
        pos.ndim == 2 and pos.shape[1] == 3), "part2se3 pos shape err"
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "part2se3 quat shape err"
    assert pos.ndim == quat.ndim, "part2se3 pos and quat must have same ndim"
    if pos.ndim == 2:
        assert pos.shape[0] == quat.shape[0], "part2se3 batch size mismatch"

    single = pos.ndim == 1
    if single:
        pos = pos[np.newaxis, :]
        quat = quat[np.newaxis, :]

    so3 = quat2so3(quat)
    se3 = np.concatenate((pos, so3), axis=1)
    return se3[0] if single else se3


def se32trans(se3: np.ndarray) -> np.ndarray:
    assert (se3.ndim == 1 and se3.shape == (6, )) or (
        se3.ndim == 2 and se3.shape[1] == 6), "se32trans se3 shape err"

    single = se3.ndim == 1
    if single:
        se3 = se3[np.newaxis, :]

    num = se3.shape[0]
    trans = np.zeros((num, 4, 4))
    trans[:, :3, 3] = se3[:, :3]
    trans[:, :3, :3] = so32rot(se3[:, 3:])
    trans[:, 3, :3] = 0.0
    trans[:, 3, 3] = 1.0
    return trans[0] if single else trans


def se32part(se3: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    assert (se3.ndim == 1 and se3.shape == (6, )) or (
        se3.ndim == 2 and se3.shape[1] == 6), "se32part se3 shape err"

    single = se3.ndim == 1
    if single:
        se3 = se3[np.newaxis, :]

    pos = se3[:, :3]
    quat = so32quat(se3[:, 3:])
    return (pos[0], quat[0]) if single else (pos, quat)


def yaw2quat(yaw) -> np.ndarray:
    if isinstance(yaw, (int, float)):
        quat = np.array([np.cos(yaw / 2), 0, 0, np.sin(yaw / 2)])
        return quat
    else:
        # Batch processing: input shape (N,) -> output shape (N, 4)
        assert isinstance(
            yaw, np.ndarray) and yaw.ndim == 1, "yaw2quat yaw shape err"
        quat = np.zeros((yaw.shape[0], 4))
        quat[:, 0] = np.cos(yaw / 2)
        quat[:, 3] = np.sin(yaw / 2)
        return quat


def quat2yaw(quat: np.ndarray):
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat2yaw quat shape err"

    single = quat.ndim == 1
    if single:
        quat = quat[np.newaxis, :]

    # Batch processing: input shape (N, 4) -> output shape (N,)
    yaw = 2 * np.arctan2(quat[:, 3], quat[:, 0])

    return yaw[0] if single else yaw


def single_euler2rot(theta: float | np.ndarray, format: str) -> np.ndarray:
    assert format in ['x', 'y',
                      'z'], f"single_euler_rot format '{format}' not supported"

    single = np.isscalar(theta)
    if single:
        theta = np.array([theta])

    rot = np.zeros((theta.shape[0], 3, 3))
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)
    if format == 'x':
        rot[:, 0, 0] = 1
        rot[:, 1, 1] = cos_theta
        rot[:, 1, 2] = -sin_theta
        rot[:, 2, 1] = sin_theta
        rot[:, 2, 2] = cos_theta
    elif format == 'y':
        rot[:, 0, 0] = cos_theta
        rot[:, 0, 2] = sin_theta
        rot[:, 1, 1] = 1
        rot[:, 2, 0] = -sin_theta
        rot[:, 2, 2] = cos_theta
    elif format == 'z':
        rot[:, 0, 0] = cos_theta
        rot[:, 0, 1] = -sin_theta
        rot[:, 1, 0] = sin_theta
        rot[:, 1, 1] = cos_theta
        rot[:, 2, 2] = 1

    return rot[0] if single else rot


def single_rot2euler(rot: np.ndarray, format: str) -> np.ndarray:
    assert format in ['x', 'y',
                      'z'], f"single_rot2euler format '{format}' not supported"
    assert (rot.ndim == 2 and rot.shape == (3, 3)) or (
        rot.ndim == 3
        and rot.shape[1:] == (3, 3)), "single_rot2euler rot shape err"
    single = rot.ndim == 2
    if single:
        rot = rot[np.newaxis, ...]

    theta = np.zeros((rot.shape[0], 3))
    if format == 'x':
        theta[:, 0] = np.arctan2(rot[:, 2, 1], rot[:, 1, 1])
    elif format == 'y':
        theta[:, 1] = np.arctan2(rot[:, 0, 2], rot[:, 0, 0])
    elif format == 'z':
        theta[:, 2] = np.arctan2(rot[:, 1, 0], rot[:, 0, 0])

    return theta[0] if single else theta


def euler2rot(euler: np.ndarray, format: str = 'xyx') -> np.ndarray:
    """
    format: 'xyx', 'xyz', 'xzx', 'xzy', 'yxy', 'yxz', 'yzx', 'yzy', 'zxy', 'zxz', 'zyx', 'zyz'
    """
    format = format.lower()
    assert format in [
        'xyx', 'xyz', 'xzx', 'xzy', 'yxy', 'yxz', 'yzx', 'yzy', 'zxy', 'zxz',
        'zyx', 'zyz'
    ], f"euler2rot format '{format}' not supported"

    assert (euler.ndim == 1 and euler.shape == (3, )) or (
        euler.ndim == 2 and euler.shape[1] == 3), "euler2rot euler shape err"

    single = euler.ndim == 1
    if single:
        euler = euler[np.newaxis, :]
    theta1, theta2, theta3 = euler[:, 0], euler[:, 1], euler[:, 2]

    rot_1 = single_euler2rot(theta1, format[0])
    rot_2 = single_euler2rot(theta2, format[1])
    rot_3 = single_euler2rot(theta3, format[2])
    result = rot_1 @ rot_2 @ rot_3
    return result[0] if single else result


def rot2euler(
    rot: np.ndarray,
    format: str = 'xyx',
    last_euler: np.ndarray | None = None,
    eps: float = 1e-6,
) -> np.ndarray:
    """
    format: 'xyx', 'xyz', 'xzx', 'xzy', 'yxy', 'yxz', 'yzx', 'yzy', 'zxy', 'zxz', 'zyx', 'zyz'
    """
    format = format.lower()
    assert format in [
        'xyx', 'xyz', 'xzx', 'xzy', 'yxy', 'yxz', 'yzx', 'yzy', 'zxy', 'zxz',
        'zyx', 'zyz'
    ], f"rot2euler format '{format}' not supported"

    assert (rot.ndim == 2 and rot.shape == (3, 3)) or (
        rot.ndim == 3 and rot.shape[1:] == (3, 3)), "rot2euler rot shape err"
    single = rot.ndim == 2
    if single:
        rot = rot[np.newaxis, ...]
    rot_num = rot.shape[0]
    if last_euler is not None:
        if single:
            assert last_euler.ndim == 1 and last_euler.shape == (
                3, ), "rot2euler last_euler shape err"
            last_euler = last_euler[np.newaxis, :]
        else:
            assert last_euler.ndim == 2 and last_euler.shape[
                1] == 3, "rot2euler last_euler shape err"
            assert last_euler.shape[
                0] == rot_num, "rot2euler last_euler and rot must have same batch size"
    else:
        last_euler = np.zeros((rot_num, 3))

    theta = last_euler.copy()

    # Proper Euler angles
    if format == 'xyx':
        theta[:, 1] = np.arccos(np.clip(rot[:, 0, 0], -1.0, 1.0))
        singular = np.abs(np.sin(theta[:, 1])) < eps
        # Non-singular cases (vectorized)
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(rot[:, 1, 0], -rot[:, 2, 0]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(rot[:, 0, 1], rot[:, 0, 2]))
        # Singular cases (loop only over singular points)
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            theta13 = np.arctan2(rot[i, 2, 1], rot[i, 1, 1])
            theta[i, 2] = theta13 - theta[i, 0]

    elif format == 'xzx':
        theta[:, 1] = np.arccos(np.clip(rot[:, 0, 0], -1.0, 1.0))
        singular = np.abs(np.sin(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(rot[:, 2, 0], rot[:, 1, 0]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(rot[:, 0, 2], -rot[:, 0, 1]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            theta13 = np.arctan2(rot[i, 2, 1], rot[i, 1, 1])
            theta[i, 2] = theta13 - theta[i, 0]

    elif format == 'yxy':
        theta[:, 1] = np.arccos(np.clip(rot[:, 1, 1], -1.0, 1.0))
        singular = np.abs(np.sin(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(rot[:, 0, 1], rot[:, 2, 1]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(rot[:, 1, 0], -rot[:, 1, 2]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            theta13 = np.arctan2(rot[i, 0, 2], rot[i, 0, 0])
            theta[i, 2] = theta13 - theta[i, 0]

    elif format == 'yzy':
        theta[:, 1] = np.arccos(np.clip(rot[:, 1, 1], -1.0, 1.0))
        singular = np.abs(np.sin(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(rot[:, 2, 1], -rot[:, 0, 1]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(rot[:, 1, 2], rot[:, 1, 0]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            theta13 = np.arctan2(rot[i, 0, 2], rot[i, 0, 0])
            theta[i, 2] = theta13 - theta[i, 0]

    elif format == 'zxz':
        theta[:, 1] = np.arccos(np.clip(rot[:, 2, 2], -1.0, 1.0))
        singular = np.abs(np.sin(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(rot[:, 0, 2], -rot[:, 1, 2]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(rot[:, 2, 0], rot[:, 2, 1]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            theta13 = np.arctan2(rot[i, 1, 0], rot[i, 0, 0])
            theta[i, 2] = theta13 - theta[i, 0]

    elif format == 'zyz':
        theta[:, 1] = np.arccos(np.clip(rot[:, 2, 2], -1.0, 1.0))
        singular = np.abs(np.sin(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(rot[:, 1, 2], rot[:, 0, 2]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(rot[:, 2, 1], -rot[:, 2, 0]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            theta13 = np.arctan2(rot[i, 1, 0], rot[i, 0, 0])
            theta[i, 2] = theta13 - theta[i, 0]

    # Tait-Bryan angles
    elif format == 'xyz':
        theta[:, 1] = np.arcsin(np.clip(rot[:, 0, 2], -1.0, 1.0))
        singular = np.abs(np.cos(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(-rot[:, 1, 2], rot[:, 2, 2]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(-rot[:, 0, 1], rot[:, 0, 0]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            if theta[i, 1] > 1.5:
                theta13 = np.arctan2(rot[i, 1, 0], rot[i, 1, 1])
                theta[i, 2] = theta13 - theta[i, 0]
            elif theta[i, 1] < -1.5:
                theta3_1 = np.arctan2(rot[i, 1, 0], rot[i, 1, 1])
                theta[i, 2] = theta3_1 + theta[i, 0]

    elif format == 'xzy':
        theta[:, 1] = np.arcsin(np.clip(-rot[:, 0, 1], -1.0, 1.0))
        singular = np.abs(np.cos(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(rot[:, 2, 1], rot[:, 1, 1]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(rot[:, 0, 2], rot[:, 0, 0]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            if theta[i, 1] > 1.5:
                theta3_1 = np.arctan2(rot[i, 1, 2], rot[i, 1, 0])
                theta[i, 2] = theta3_1 + theta[i, 0]
            elif theta[i, 1] < -1.5:
                theta13 = np.arctan2(-rot[i, 1, 2], -rot[i, 1, 0])
                theta[i, 2] = theta13 - theta[i, 0]

    elif format == 'yxz':
        theta[:, 1] = np.arcsin(np.clip(-rot[:, 1, 2], -1.0, 1.0))
        singular = np.abs(np.cos(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(rot[:, 0, 2], rot[:, 2, 2]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(rot[:, 1, 0], rot[:, 1, 1]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            if theta[i, 1] > 1.5:
                theta3_1 = np.arctan2(rot[i, 2, 0], rot[i, 2, 1])
                theta[i, 2] = theta3_1 + theta[i, 0]
            elif theta[i, 1] < -1.5:
                theta13 = np.arctan2(-rot[i, 2, 0], -rot[i, 2, 1])
                theta[i, 2] = theta13 - theta[i, 0]

    elif format == 'yzx':
        theta[:, 1] = np.arcsin(np.clip(rot[:, 1, 0], -1.0, 1.0))
        singular = np.abs(np.cos(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(-rot[:, 2, 0], rot[:, 0, 0]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(-rot[:, 1, 2], rot[:, 1, 1]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            if theta[i, 1] > 1.5:
                theta13 = np.arctan2(rot[i, 2, 1], rot[i, 2, 2])
                theta[i, 2] = theta13 - theta[i, 0]
            elif theta[i, 1] < -1.5:
                theta3_1 = np.arctan2(rot[i, 2, 1], rot[i, 2, 2])
                theta[i, 2] = theta3_1 + theta[i, 0]

    elif format == 'zxy':
        theta[:, 1] = np.arcsin(np.clip(rot[:, 2, 1], -1.0, 1.0))
        singular = np.abs(np.cos(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(-rot[:, 0, 1], rot[:, 1, 1]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(-rot[:, 2, 0], rot[:, 2, 2]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            if theta[i, 1] > 1.5:
                theta13 = np.arctan2(rot[i, 1, 0], rot[i, 0, 0])
                theta[i, 2] = theta13 - theta[i, 0]
            elif theta[i, 1] < -1.5:
                theta3_1 = np.arctan2(-rot[i, 1, 0], rot[i, 0, 0])
                theta[i, 2] = theta3_1 + theta[i, 0]

    elif format == 'zyx':
        theta[:, 1] = np.arcsin(np.clip(-rot[:, 2, 0], -1.0, 1.0))
        singular = np.abs(np.cos(theta[:, 1])) < eps
        theta[:, 0] = np.where(singular, theta[:, 0],
                               np.arctan2(rot[:, 1, 0], rot[:, 0, 0]))
        theta[:, 2] = np.where(singular, theta[:, 2],
                               np.arctan2(rot[:, 2, 1], rot[:, 2, 2]))
        singular_idx = np.where(singular)[0]
        for i in singular_idx:
            if theta[i, 1] > 1.5:
                theta3_1 = np.arctan2(rot[i, 0, 1], rot[i, 1, 1])
                theta[i, 2] = theta3_1 + theta[i, 0]
            elif theta[i, 1] < -1.5:
                theta13 = np.arctan2(-rot[i, 0, 1], rot[i, 1, 1])
                theta[i, 2] = theta13 - theta[i, 0]

    return theta[0] if single else theta



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-11-18
################################################################

import numpy as np
import pytest

from hex_robo_utils import math_utils as mu


def test_hat_and_vee_single_and_batch():
    vec = np.array([1.0, 2.0, 3.0])
    expected = np.array([
        [0.0, -3.0, 2.0],
        [3.0, 0.0, -1.0],
        [-2.0, 1.0, 0.0],
    ])
    assert np.allclose(mu.hat(vec), expected)

    batch_vec = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
    mats = mu.hat(batch_vec)
    assert mats.shape == (2, 3, 3)
    assert np.allclose(mu.vee(mats), batch_vec)


def test_angle_conversions_and_norm():
    assert mu.rad2deg(np.pi) == pytest.approx(180.0)
    assert mu.deg2rad(180.0) == pytest.approx(np.pi)
    assert mu.angle_norm(3 * np.pi) == pytest.approx(-np.pi)


def test_vec_and_quat_norm():
    vec = np.array([3.0, 4.0, 0.0])
    assert np.allclose(mu.vec_norm(vec), np.array([0.6, 0.8, 0.0]))

    quat = np.array([-2.0, 0.0, 0.0, 0.0])
    normed = mu.quat_norm(quat)
    assert np.allclose(normed, np.array([1.0, 0.0, 0.0, 0.0]))


def test_quat_slerp_midpoint():
    q1 = np.array([1.0, 0.0, 0.0, 0.0])
    q2 = np.array([0.0, 0.0, 0.0, 1.0])
    result = mu.quat_slerp(q1, q2, 0.5)
    expected = np.array([np.sqrt(0.5), 0.0, 0.0, np.sqrt(0.5)])
    assert np.allclose(result, expected, atol=1e-6)


def test_quat_mul_and_inv_identity():
    yaw_quat = mu.yaw2quat(np.pi / 2)
    inv = mu.quat_inv(yaw_quat)
    identity = mu.quat_mul(yaw_quat, inv)
    assert np.allclose(identity, np.array([1.0, 0.0, 0.0, 0.0]), atol=1e-6)


def test_trans_inv_round_trip():
    pos = np.array([1.0, -2.0, 0.5])
    rot = mu.axis2rot(np.array([0.0, 0.0, 1.0]), np.pi / 4)
    trans = np.eye(4)
    trans[:3, :3] = rot
    trans[:3, 3] = pos

    inv = mu.trans_inv(trans)
    identity = trans @ inv
    assert np.allclose(identity, np.eye(4), atol=1e-6)


def test_rotation_quaternion_conversions():
    axis = np.array([0.0, 0.0, 1.0])
    theta = np.pi / 3
    rot = mu.axis2rot(axis, theta)
    quat = mu.rot2quat(rot)
    recon_rot = mu.quat2rot(quat)
    assert np.allclose(rot, recon_rot, atol=1e-6)


def test_rot_axis_so3_conversions():
    axis = np.array([0.0, 1.0, 0.0])
    theta = 0.3
    rot = mu.axis2rot(axis, theta)
    axis_out, theta_out = mu.rot2axis(rot)
    assert np.allclose(theta_out, theta, atol=1e-6)
    assert np.allclose(axis_out, axis, atol=1e-6)

    so3 = mu.rot2so3(rot)
    rot_from_so3 = mu.so32rot(so3)
    assert np.allclose(rot_from_so3, rot, atol=1e-6)


def test_quaternion_axis_so3_conversions():
    axis = np.array([1.0, 0.0, 0.0])
    theta = 0.4
    quat = mu.axis2quat(axis, theta)
    axis_out, theta_out = mu.quat2axis(quat)
    assert np.allclose(theta_out, theta, atol=1e-6)
    assert np.allclose(axis_out, axis, atol=1e-6)

    so3 = mu.quat2so3(quat)
    quat_from_so3 = mu.so32quat(so3)
    assert np.allclose(quat_from_so3, mu.quat_norm(quat), atol=1e-6)

    axis_from_so3, theta_from_so3 = mu.so32axis(so3)
    assert np.allclose(theta_from_so3, theta, atol=1e-6)
    assert np.allclose(axis_from_so3, axis, atol=1e-6)


def test_axis2so3_and_vec_norm_integration():
    axis = np.array([0.0, 2.0, 0.0])
    theta = 0.25
    so3 = mu.axis2so3(axis, theta)
    assert np.allclose(so3, np.array([0.0, 0.25, 0.0]), atol=1e-6)


def test_transform_and_se3_conversions():
    pos = np.array([0.2, -0.1, 0.3])
    quat = mu.axis2quat(np.array([0.0, 0.0, 1.0]), 0.2)
    trans = mu.part2trans(pos, quat)

    pos_out, quat_out = mu.trans2part(trans)
    assert np.allclose(pos_out, pos, atol=1e-6)
    assert np.allclose(quat_out, mu.quat_norm(quat), atol=1e-6)

    se3 = mu.trans2se3(trans)
    trans_from_se3 = mu.se32trans(se3)
    assert np.allclose(trans_from_se3, trans, atol=1e-6)

    se3_from_part = mu.part2se3(pos, quat)
    pos_se3, quat_se3 = mu.se32part(se3_from_part)
    assert np.allclose(pos_se3, pos, atol=1e-6)
    assert np.allclose(quat_se3, mu.quat_norm(quat), atol=1e-6)


def test_yaw_quaternion_conversion():
    yaw = 0.7
    quat = mu.yaw2quat(yaw)
    recovered = mu.quat2yaw(quat)
    assert recovered == pytest.approx(yaw, abs=1e-6)


def test_single_euler_helpers():
    theta = np.pi / 6
    rot_x = mu.single_euler2rot(theta, 'x')
    recovered = mu.single_rot2euler(rot_x, 'x')
    assert recovered[0] == pytest.approx(theta, abs=1e-6)

    # Vector input path
    theta_vec = np.array([theta, theta])
    rot_batch = mu.single_euler2rot(theta_vec, 'z')
    assert rot_batch.shape == (2, 3, 3)


def test_euler_rot_round_trip():
    euler = np.array([0.1, -0.2, 0.3])
    rot = mu.euler2rot(euler, format='xyz')
    recovered = mu.rot2euler(rot, format='xyz')
    assert np.allclose(recovered, euler, atol=1e-6)


def test_twist_swing_decomposition():
    twist_true = mu.axis2quat(np.array([0.0, 0.0, 1.0]), 0.3)
    swing_true = mu.axis2quat(np.array([1.0, 0.0, 0.0]), 0.4)
    quat = mu.quat_mul(swing_true, twist_true)

    twist, swing = mu.twist_swing_decomp(quat, np.array([0.0, 0.0, 1.0]))
    assert np.allclose(twist, mu.quat_norm(twist_true), atol=1e-6)
    assert np.allclose(swing, mu.quat_norm(swing_true), atol=1e-6)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-12-04
################################################################

import os
import time
import multiprocessing
import threading
import numpy as np

try:
    from hex_robo_utils.hdf5_writer import HexHdf5MultiWriter
except ImportError:
    import sys
    sys.path.insert(
        0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    from hex_robo_utils.hdf5_writer import HexHdf5MultiWriter


class HexRate:

    def __init__(self, hz: float, spin_threshold_ns: int = 10_000):
        if hz <= 0:
            raise ValueError("hz must be greater than 0")
        if spin_threshold_ns < 0:
            raise ValueError("spin_threshold_ns must be non-negative")
        self.__period_ns = int(1_000_000_000 / hz)
        self.__next_ns = self.__now_ns() + self.__period_ns
        self.__spin_threshold_ns = spin_threshold_ns

    @staticmethod
    def __now_ns() -> int:
        return time.perf_counter_ns()

    def reset(self):
        self.__next_ns = self.__now_ns() + self.__period_ns

    def sleep(self):
        target_ns = self.__next_ns
        now_ns = self.__now_ns()
        remain_ns = target_ns - now_ns
        if remain_ns <= 0:
            needed_period = (now_ns - target_ns) // self.__period_ns + 1
            self.__next_ns += needed_period * self.__period_ns
            return

        spin_threshold = min(self.__spin_threshold_ns, self.__period_ns)
        coarse_sleep_ns = remain_ns - spin_threshold
        if coarse_sleep_ns > 0:
            time.sleep(coarse_sleep_ns / 1_000_000_000.0)

        while True:
            now_ns = self.__now_ns()
            if now_ns >= target_ns:
                break
            if target_ns - now_ns > 50_000:
                time.sleep(0)

        self.__next_ns += self.__period_ns


class MultiArmRGBDRecorder:

    def __init__(
        self,
        base_dir: str,
        duration_s: float = 30.0,
        num_arms: int = 6,
        num_cams: int = 4,
        arm_hz: int = 1000,
        cam_hz: int = 30,
    ):

        self.duration_ns = int(duration_s * 1_000_000_000)
        self.num_arms = num_arms
        self.num_cams = num_cams
        self.arm_hz = arm_hz
        self.cam_hz = cam_hz

        # 使用多文件 writer，将不同类型数据写入不同的 h5 文件
        self._writer = HexHdf5MultiWriter(base_dir)

        self.arm_shape = (7, 3)
        self.arm_dtype = np.float64
        self.rgb_shape = (480, 640, 3)
        self.rgb_dtype = np.uint8
        self.depth_shape = (480, 640)
        self.depth_dtype = np.uint16

        self._processes: list[multiprocessing.Process] = []
        self._stop_event = multiprocessing.Event()
        self._start_time_ns: int | None = None
        self._manager = None
        self._data_queue = None
        self._queue_thread = None

        self._create_datasets()

    # ----------------------- public API -----------------------

    def start(self):
        if self._start_time_ns is not None:
            return
        self._stop_event.clear()
        self._writer.start()
        # 统一使用 perf_counter_ns 作为时间基准
        self._start_time_ns = time.perf_counter_ns()

        # 创建 Manager 用于进程间共享
        self._manager = multiprocessing.Manager()
        shared_start_time = self._manager.Value('i', self._start_time_ns)
        self._data_queue = self._manager.Queue()

        # 启动队列处理线程
        self._queue_thread = threading.Thread(target=self._queue_worker,
                                              daemon=True)
        self._queue_thread.start()

        for arm_id in range(self.num_arms):
            p = multiprocessing.Process(
                target=self._arm_process,
                args=(arm_id, self._stop_event, shared_start_time,
                      self.duration_ns, self.arm_hz, self.arm_shape,
                      self.arm_dtype, self._data_queue))
            self._processes.append(p)
            p.start()

        for cam_id in range(self.num_cams):
            p = multiprocessing.Process(
                target=self._rgbd_process,
                args=(cam_id, self._stop_event, shared_start_time,
                      self.duration_ns, self.cam_hz, self.rgb_shape,
                      self.rgb_dtype, self.depth_shape, self.depth_dtype,
                      self._data_queue))
            self._processes.append(p)
            p.start()

    def wait(self):
        if self._start_time_ns is None:
            return

        end_time = self._start_time_ns + self.duration_ns
        while (not self._stop_event.is_set()
               ) and time.perf_counter_ns() < end_time:
            time.sleep(0.1)

        self.stop()

    def stop(self):
        """停止所有进程并关闭 writer。"""
        if self._start_time_ns is None:
            return

        self._stop_event.set()
        for p in self._processes:
            if p.is_alive():
                p.join()
        self._processes.clear()

        # 等待队列处理完成
        if self._data_queue is not None:
            # 发送结束标记
            for _ in range(self.num_arms + self.num_cams):
                self._data_queue.put(None)
            if self._queue_thread is not None and self._queue_thread.is_alive(
            ):
                self._queue_thread.join(timeout=5.0)

        self._writer.stop()
        self._start_time_ns = None
        if self._manager is not None:
            self._manager.shutdown()
            self._manager = None
        self._data_queue = None

    def run(self):
        self.start()
        self.wait()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    # --------------------- internal helpers --------------------

    def _create_datasets(self):
        for arm_id in range(self.num_arms):
            group = f"arm_{arm_id}"
            self._writer.create_dataset(
                "arm",
                group,
                shape=self.arm_shape,
                dtype=self.arm_dtype,
                chunk_num=1024,
                max_num=None,
            )

        for cam_id in range(self.num_cams):
            rgb_group = f"cam_{cam_id}_rgb"
            depth_group = f"cam_{cam_id}_depth"

            self._writer.create_dataset(
                "rgb",
                rgb_group,
                shape=self.rgb_shape,
                dtype=self.rgb_dtype,
                chunk_num=1,
                max_num=None,
            )
            self._writer.create_dataset(
                "depth",
                depth_group,
                shape=self.depth_shape,
                dtype=self.depth_dtype,
                chunk_num=1,
                max_num=None,
            )

    def _queue_worker(self):
        """从队列中读取数据并写入 HDF5"""
        import queue
        while True:
            try:
                item = self._data_queue.get(timeout=1.0)
                if item is None:  # 结束标记
                    continue
                msg_type, group, data, get_ts, sen_ts = item
                self._writer.append_data(msg_type, group, data, get_ts, sen_ts)
            except queue.Empty:
                # 超时，检查是否应该退出
                if self._stop_event.is_set():
                    # 处理剩余数据
                    while True:
                        try:
                            item = self._data_queue.get_nowait()
                            if item is None:
                                continue
                            msg_type, group, data, get_ts, sen_ts = item
                            self._writer.append_data(msg_type, group, data,
                                                     get_ts, sen_ts)
                        except queue.Empty:
                            break
                    break
            except Exception as e:
                print(f"Queue worker error: {e}")
                if self._stop_event.is_set():
                    break

    @staticmethod
    def _time_remain(shared_start_time, duration_ns) -> bool:
        if shared_start_time.value == 0:
            return False
        return (time.perf_counter_ns() - shared_start_time.value) < duration_ns

    @staticmethod
    def _arm_process(arm_id: int, stop_event: multiprocessing.Event,
                     shared_start_time, duration_ns: int, arm_hz: int,
                     arm_shape: tuple, arm_dtype: np.dtype, data_queue):
        group = f"arm_{arm_id}"
        hex_rate = HexRate(arm_hz)
        fps_cnt = 0
        start_time_ns = time.perf_counter_ns()
        while (not stop_event.is_set()) and MultiArmRGBDRecorder._time_remain(
                shared_start_time, duration_ns):
            data = np.random.randn(*arm_shape).astype(arm_dtype)
            get_ts = time.time_ns()
            sen_ts = time.time_ns()
            # 将数据放入队列
            data_queue.put(("arm", group, data, get_ts, sen_ts))
            fps_cnt += 1
            if fps_cnt >= 3_000:
                delta_s = (time.perf_counter_ns() - start_time_ns) * 1e-9
                print(f"Arm {arm_id} FPS: {fps_cnt / delta_s}Hz")
                fps_cnt = 0
                start_time_ns = time.perf_counter_ns()
            hex_rate.sleep()

    @staticmethod
    def _rgbd_process(cam_id: int, stop_event: multiprocessing.Event,
                      shared_start_time, duration_ns: int, cam_hz: int,
                      rgb_shape: tuple, rgb_dtype: np.dtype,
                      depth_shape: tuple, depth_dtype: np.dtype, data_queue):
        rgb_group = f"cam_{cam_id}_rgb"
        depth_group = f"cam_{cam_id}_depth"
        hex_rate = HexRate(cam_hz)
        fps_cnt = 0
        start_time_ns = time.perf_counter_ns()
        while (not stop_event.is_set()) and MultiArmRGBDRecorder._time_remain(
                shared_start_time, duration_ns):
            rgb = np.random.randint(
                0,
                256,
                size=rgb_shape,
                dtype=rgb_dtype,
            )
            depth = np.random.randint(
                0,
                65536,
                size=depth_shape,
                dtype=depth_dtype,
            )
            get_ts = time.time_ns()
            sen_ts = time.time_ns()
            # 将数据放入队列
            data_queue.put(("rgb", rgb_group, rgb, get_ts, sen_ts))
            data_queue.put(("depth", depth_group, depth, get_ts, sen_ts))
            fps_cnt += 1
            if fps_cnt >= 100:
                delta_s = (time.perf_counter_ns() - start_time_ns) * 1e-9
                print(f"RGBD {cam_id} FPS: {fps_cnt / delta_s}Hz")
                fps_cnt = 0
                start_time_ns = time.perf_counter_ns()
            hex_rate.sleep()


def main():
    out_path = os.path.abspath("multi_arm_rgbd")
    print(f"Recording base: {out_path}")

    start_ns = time.perf_counter_ns()
    recorder = MultiArmRGBDRecorder(
        out_path,
        duration_s=30.0,
        num_arms=6,
        num_cams=4,
        arm_hz=1000,
        cam_hz=30,
    )
    recorder.run()
    print("#" * 50)
    print(f"Time taken: {(time.perf_counter_ns() - start_ns) * 1e-6}ms")
    print("#" * 50)
    print("Done.")


if __name__ == "__main__":
    main()